"""
Interface to connect two or more nodes (Docker containers) together using point-to-point ZMQ sockets (client-server)
"""

import os
import time

import zmq

#  Socket to talk to server
context = zmq.Context()


class ZdgNodeIface:
    """
    ZdgNodeIface
    """

    def __init__(self) -> None:
        pass

    @staticmethod
    def create_inbound_sockets():
        """
        create_client_sockets
        """

        inbound_list = str(os.environ["ZDG_INBOUND_LIST"])
        print(f"Inbound list: {inbound_list}")

        if len(inbound_list) == 0:
            return {}

        inbound_sockets = {}
        socket_cnt = 0
        for h_p in inbound_list.split(";"):
            socket_cnt += 1

            h, p = h_p.split()
            # print(f"h_p {h_p}")
            # print(f"h {h}")
            # print(f"p {p}")

            socket_hostname = str(h)
            socket_port = int(p)

            socket_socket = context.socket(zmq.REP)
            socket_url = f"tcp://{socket_hostname}:{socket_port}"
            print(f"Binding socket to {socket_url}")
            socket_socket.bind(socket_url)

            # Subscribe to zipcode, default is NYC, 10001
            # topic_filter = "10001"
            # socket_sub.setsockopt_string(zmq.SUBSCRIBE, topic_filter)
            # client_socket.setsockopt_string(zmq.SUBSCRIBE, "")

            inbound_sockets[f"socket_{socket_cnt}"] = {
                "hostname": socket_hostname,
                "port": socket_port,
                "url": socket_url,
                "socket": socket_socket,
            }

        return inbound_sockets

    @staticmethod
    def create_outbound_sockets():
        """
        create_server_sockets
        """

        outbound_list = str(os.environ["ZDG_OUTBOUND_LIST"])
        print(f"Outbound list: {outbound_list}")

        if len(outbound_list) == 0:
            return {}

        outbound_sockets = {}
        socket_cnt = 0
        for h_p in outbound_list.split(";"):
            socket_cnt += 1

            h, p = h_p.split()
            # print(f"h_p {h_p}")
            # print(f"h {h}")
            # print(f"p {p}")

            socket_hostname = str(h)
            # pub_hostname = "0.0.0.0"
            socket_port = int(p)

            # pub_socket = context.socket(zmq.PUB)
            # pub_url = f"tcp://{pub_hostname}:{pub_port}"
            # print(f"Connecting socket to {pub_url}")
            # pub_socket.connect(pub_url)

            socket_socket = context.socket(zmq.REQ)
            socket_url = f"tcp://{socket_hostname}:{socket_port}"
            print(f"Connecting socket to {socket_url}")
            socket_socket.connect(socket_url)

            outbound_sockets[f"socket_{socket_cnt}"] = {
                "hostname": socket_hostname,
                "port": socket_port,
                "url": socket_url,
                "socket": socket_socket,
            }

        return outbound_sockets

    @staticmethod
    def process_outbound_message(message: dict, socket_dict: dict):
        socket = socket_dict["socket"]
        socket_url = socket_dict["url"]

        verbose = False

        # Send request to server
        if verbose:
            print(f"[outbound] Sending to     {socket_url} a reqst with keys: {message.keys()}")
        try:
            socket.send_pyobj(message)
        except zmq.error.ZMQError:
            print("")
            print("")
            fname = ZdgNodeIface.process_outbound_message.__name__
            print(f"[{fname}] zmq.error.ZMQError: Socket operation on non-socket")
            print(f"[{fname}] {ZdgNodeIface.create_outbound_sockets.__name__}")
            print("")
            print("")
            ZdgNodeIface.create_outbound_sockets()

        # # Get the reply from server
        # reply = socket.recv_pyobj()
        # print(f"[outbound] Receiving from {socket_url}: {reply}")

        REQUEST_TIMEOUT = 2500
        REQUEST_RETRIES = 3
        SERVER_ENDPOINT = socket_url

        retries_left = REQUEST_RETRIES
        while True:
            if (socket.poll(REQUEST_TIMEOUT) & zmq.POLLIN) != 0:
                reply = socket.recv_pyobj()
                if verbose:
                    print(f"[outbound] Receiving from {socket_url} a reply with keys: {reply.keys()}")
                break
                # if int(reply) == sequence:
                #     print("Server replied OK (%s)", reply)
                #     retries_left = REQUEST_RETRIES
                #     break
                # else:
                #     print("Malformed reply from server: %s", reply)
                #     continue

            retries_left -= 1
            print("No response from server")
            # Socket is confused. Close and remove it.
            socket.setsockopt(zmq.LINGER, 0)
            socket.close()
            if retries_left == 0:
                print("Server seems to be offline, abandoning")
                raise RuntimeError

            print("Reconnecting to server")
            # Create new connection
            socket = context.socket(zmq.REQ)
            socket.connect(SERVER_ENDPOINT)
            print("Resending (%s)", message.keys())
            socket.send_pyobj(message)

        return message, reply

    @staticmethod
    def process_inbound_message(inbound_fnct, socket_dict: dict):
        socket = socket_dict["socket"]
        socket_url = socket_dict["url"]

        verbose = False

        # Wait for the next request from client
        message = socket.recv_pyobj()
        if verbose:
            print(f"[inbound]  Receiving from {socket_url}     a reqst with keys: {message.keys()}")

        # Process message
        reply = inbound_fnct(message)
        assert isinstance(reply, dict)

        # Send reply to client
        if verbose:
            print(f"[inbound]  Sending to     {socket_url}     a reply with keys: {reply.keys()}")
        socket.send_pyobj(reply)

        return message, reply

    @staticmethod
    def acknowledge_message(message):
        reply = {"message_time": message["time"]}
        return reply

    @staticmethod
    def process_m_to_0_communication(inbound_sockets: dict, inbound_fnct):
        """
        process_m_to_0_communication
        """
        fname = ZdgNodeIface.process_m_to_0_communication.__name__
        print(fname)

        # https://zguide.zeromq.org/docs/chapter2/#Handling-Multiple-Sockets

        # Initialize poll set
        poller = zmq.Poller()

        for in_key, in_val in inbound_sockets.items():
            in_socket_cnt = in_key
            in_socket_socket = in_val["socket"]
            in_socket_url = in_val["url"]
            print(f"Register socket to Poller: socket_cnt {in_socket_cnt}, url {in_socket_url}")
            poller.register(in_socket_socket, zmq.POLLIN)

        # Process messages from both sockets
        message_t0 = time.time()
        message_cnt = 0
        while True:
            message_cnt += 1

            try:
                socket_list = dict(poller.poll())
            except KeyboardInterrupt:
                break

            for in_key, in_val in inbound_sockets.items():
                in_socket_cnt = in_key
                in_socket_socket = in_val["socket"]
                in_socket_url = in_val["url"]

                if in_socket_socket in socket_list:
                    message, reply = ZdgNodeIface.process_inbound_message(inbound_fnct=inbound_fnct, socket_dict=in_val)
                    _ = message
                    _ = reply

                    if message_cnt % 100 == 0:
                        message_t100 = time.time()
                        message_dt = message_t100 - message_t0
                        message_rate = message_cnt / message_dt
                        print(f"[{fname}] Effective message rate {message_rate} message/s")
                        message_t0 = time.time()

    @staticmethod
    def process_0_to_n_communication(outbound_sockets: dict, outbound_fnct):
        """
        process_0_to_n_communication
        """
        fname = ZdgNodeIface.process_0_to_n_communication.__name__
        print(fname)

        # This is not an ideal solution
        # Allow for downstream nodes to initalize
        time.sleep(10)

        # Process messages from both sockets
        message_t0 = time.time()
        message_cnt = 0
        while True:
            message_cnt += 1

            message = {
                "data": outbound_fnct(),
                "time": time.time(),
                "counter": message_cnt,
            }

            for out_key, out_val in outbound_sockets.items():
                # out_socket_cnt = out_key
                # out_socket_socket = out_val["socket"]
                # out_socket_url = out_val["url"]
                _ = out_key

                ZdgNodeIface.process_outbound_message(message=message, socket_dict=out_val)

                if message_cnt % 100 == 0:
                    message_t100 = time.time()
                    message_dt = message_t100 - message_t0
                    message_rate = message_cnt / message_dt
                    print(f"[{fname}] Effective message rate {message_rate} message/s")
                    message_t0 = time.time()

            # time.sleep(1/15)
            time.sleep(60)

    @staticmethod
    def process_n_to_m_communication(inbound_sockets: dict, inbound_fnct, outbound_sockets: dict, outbound_fnct):
        """
        process_n_to_m_communication
        """
        print(ZdgNodeIface.process_n_to_m_communication.__name__)

        # https://zguide.zeromq.org/docs/chapter2/#Handling-Multiple-Sockets

        # Initialize poll set
        poller = zmq.Poller()

        # # Initialize poll set
        # poller = zmq.Poller()
        # poller.register(receiver, zmq.POLLIN)
        # poller.register(subscriber, zmq.POLLIN)

        # # Process messages from both sockets
        # while True:
        #     try:
        #         socks = dict(poller.poll())
        #     except KeyboardInterrupt:
        #         break

        #     if receiver in socks:
        #         message = receiver.recv_pyobj()
        #         # process task

        #     if subscriber in socks:
        #         message = subscriber.recv_pyobj()
        #         # process weather update

        for in_key, in_val in inbound_sockets.items():
            in_socket_cnt = in_key
            in_socket_socket = in_val["socket"]
            in_socket_url = in_val["url"]
            print(f"Register socket to Poller: socket_cnt {in_socket_cnt}, url {in_socket_url}")
            poller.register(in_socket_socket, zmq.POLLIN)

        # Process messages from both sockets
        message_cnt = 0
        while True:
            message_cnt += 1

            try:
                socket_list = dict(poller.poll())
            except KeyboardInterrupt:
                break

            # if receiver in active_socket:
            #     message = receiver.recv_pyobj()
            #     # process task

            # if subscriber in active_socket:
            #     message = subscriber.recv_pyobj()
            #     # process weather update

            for in_key, in_val in inbound_sockets.items():
                in_socket_cnt = in_key
                in_socket_socket = in_val["socket"]
                in_socket_url = in_val["url"]

                if in_socket_socket in socket_list:
                    inbound_message, inbound_reply = ZdgNodeIface.process_inbound_message(
                        # target_fnct=ZmqNodeIface.acknowledge_message,
                        inbound_fnct=inbound_fnct,
                        socket_dict=in_val,
                    )
                    _ = inbound_reply

                    # For every inbound message, send the same outbound message data to all outbound sockets
                    outbound_data = outbound_fnct(inbound_message)

                    for out_key, out_val in outbound_sockets.items():
                        # out_socket_cnt = out_key
                        # out_socket_socket = out_val["socket"]
                        # out_socket_url = out_val["url"]
                        _ = out_key

                        outbound_message = {
                            "data": outbound_data,
                            "time": time.time(),
                            "counter": message_cnt,
                        }

                        ZdgNodeIface.process_outbound_message(message=outbound_message, socket_dict=out_val)

    @staticmethod
    def run(inbound_fnct, outbound_fnct):
        """
        Pull data from subscribers and push data to publishers
        """

        # print(f"os.uname() {os.uname()}")
        # print(f"print(os.environ) {os.environ}")
        print(f"zmq.zmq_version() {zmq.zmq_version()}")
        # print(f"zmq.__version__ {zmq.__version__}")

        inbound_sockets = ZdgNodeIface.create_inbound_sockets()
        outbound_sockets = ZdgNodeIface.create_outbound_sockets()

        n_in = len(inbound_sockets)
        m_out = len(outbound_sockets)

        print(f"inbound length {n_in}")
        print(f"outbound length {m_out}")

        if (n_in > 0) and (m_out > 0):
            ZdgNodeIface.process_n_to_m_communication(inbound_sockets, inbound_fnct, outbound_sockets, outbound_fnct)
        elif (n_in == 0) and (m_out > 0):
            ZdgNodeIface.process_0_to_n_communication(outbound_sockets, outbound_fnct)
        elif (n_in > 0) and (m_out == 0):
            ZdgNodeIface.process_m_to_0_communication(inbound_sockets, inbound_fnct)
        elif (n_in == 0) and (m_out == 0):
            print("Nothing to do here, empty inbound_sockets and outbound_sockets")
        else:
            raise RuntimeError



"""
Interface to connect two or more nodes (Docker containers) together using point-to-point ZMQ sockets (client-server)
"""

import os

import zmq

from zdg.node_interface import ZdgNodeIface

# import pathlib
# import time


#  Socket to talk to server
context = zmq.Context()


class ZdgNodeDemo:
    """
    ZdgNodeDemo
    """

    def __init__(self) -> None:
        self.zmq_container_name = str(os.environ["ZDG_CONTAINER_NAME"])

    def outbound_fnct(self, message={}):
        """
        outbound_fnct should be passed to ZmqNodeIface.run function and then be
        internally called by ZmqNodeIface functions
          process_n_to_m_communication
          process_0_to_n_communication
        """
        if len(message) == 0:
            # print("Assuming 0 to n role")

            data = f"From {self.zmq_container_name} with love"

            print(f"Outbound message data    {data}")
            # print(f"Outbound message time    {time}")
            # print(f"Outbound message counter {counter}")

            return data

        # print("Assuming n to m role")
        data = message["data"]
        # time = message["time"]
        # counter = message["counter"]
        # print(f"Inbound message data     {data}")
        # print(f"Inbound message time     {time}")
        # print(f"Inbound message counter  {counter}")

        data = f"{data} > From {self.zmq_container_name} with love"

        print(f"Outbound message data    {data}")
        # print(f"Outbound message time    {time}")
        # print(f"Outbound message counter {counter}")

        return data

    def inbound_fnct(self, message={}):
        """
        inbound_fnct should be passed to ZmqNodeIface.run function and then be
        internally called by ZmqNodeIface functions
          process_n_to_m_communication
          process_m_to_0_communication
        """
        # print("Assuming n to m (m>=0) role")
        mdata = message["data"]
        mtime = message["time"]
        mcounter = message["counter"]
        print(f"Inbound message data     {mdata}")
        print(f"Inbound message time     {mtime}")
        print(f"Inbound message counter  {mcounter}")

        reply = {"reply": f"Reply from node {self.zmq_container_name} to message {mcounter}"}
        return reply


if __name__ == "__main__":
    znd = ZdgNodeDemo()
    ZdgNodeIface.run(inbound_fnct=znd.inbound_fnct, outbound_fnct=znd.outbound_fnct)



"""
Create a directed graph where each node is a Docker container and each edge is a ZMQ point-to-point socket
"""
import os
import pathlib

import yaml

# import random


class ZdgNode:
    """
    Class to create a directed graph where each node is a Docker container and each edge is a ZMQ point-to-point socket
    """

    def __init__(self, container_name: str, image: str, command: str) -> None:
        """
        __init__
        """
        self.current_dir = pathlib.Path(__file__).parent.resolve()

        self.container_name = container_name
        self.image = image
        self.command = command

        # self.volumes = [".:/app"]
        self.volumes = []
        self.environment = [f"ZDG_CONTAINER_NAME={self.container_name}"]
        self.depends_on = []

    def update_inbound_list(self, in_h_list: list, in_p_list: list):
        """
        update_inbound_list
        """
        env = "ZDG_INBOUND_LIST"
        h_p_list = ""
        for _, in_p in zip(in_h_list, in_p_list):
            if len(h_p_list) == 0:
                # h_p_list = f"{in_h} {port}"
                h_p_list = f"* {in_p}"
            else:
                # h_p_list = f"{h_p_list};{in_h} {port}"
                h_p_list = f"{h_p_list};* {in_p}"
        self.environment.append(f"{env}={h_p_list}")

        self.depends_on = in_h_list

    def update_outbound_list(self, out_h_list: list, out_p_list: list):
        """
        update_outbound_list
        """
        env = "ZDG_OUTBOUND_LIST"
        h_p_list = ""
        for out_h, out_p in zip(out_h_list, out_p_list):
            if len(h_p_list) == 0:
                h_p_list = f"{out_h} {out_p}"
            else:
                h_p_list = f"{h_p_list};{out_h} {out_p}"
        self.environment.append(f"{env}={h_p_list}")

    def update_yml(self):
        """
        update_yml
        """
        datad = {
            self.container_name: {
                "command": self.command,
                "container_name": self.container_name,
                "depends_on": self.depends_on,
                "environment": self.environment,
                "image": self.image,
                "volumes": self.volumes,
                # "working_dir": self.working_dir
            }
        }
        return datad

    @staticmethod
    def write_yml(datad: dict, yaml_path: str):
        """
        write_yml
        """
        # datad = self.update_yml()
        # yaml_path = str(self.current_dir / f"pubsub_compose_{self.container_name}.yml")
        with open(yaml_path, "w", encoding="utf-8") as f_d:
            yaml.dump(datad, f_d, default_flow_style=False)

    @staticmethod
    def get_comments():
        """
        get_comments
        """
        l1 = "# Run it using"
        l2 = "#   sudo docker compose -f compose_dgraph.yml up --remove-orphans"
        l3 = ""
        l4 = "# Execute a commnad on a running container using"
        l5 = "#   docker exec -it zdg bash"
        l6 = ""
        l7 = "# Remove all stopped containers"
        l8 = "#   sudo docker rm $(sudo docker ps --filter status=exited -q)"
        return [l1, l2, l3, l4, l5, l6, l7, l8]

    @staticmethod
    def demo_compose():
        """
        demo_compose
        """
        compose_data = {"services": {}}

        current_dir = pathlib.Path(__file__).parent.resolve()

        node_demo_command = "/bin/bash node_demo.sh"
        working_dir = "/demo"

        num_source_nodes = 2
        num_middle_nodes = 1  # random.randint(1, 5)
        num_sink_nodes = 1

        # Creating nodes
        source_node_list = []
        for i in range(0, num_source_nodes):
            container_name = f"source_node_{i}"
            source_ni = ZdgNode(container_name, "zdg", node_demo_command)
            source_ni.volumes.append(f"{os.path.split(__file__)[0]}:{working_dir}")
            source_node_list.append(source_ni)
            print(f"Creating source node {container_name}")

        middle_node_list = []
        for i in range(0, num_middle_nodes):
            container_name = f"middle_node_{i}"
            middle_ni = ZdgNode(container_name, "zdg", node_demo_command)
            middle_ni.volumes.append(f"{os.path.split(__file__)[0]}:{working_dir}")
            middle_node_list.append(middle_ni)
            print(f"Creating middle node {container_name}")

        sink_node_list = []
        for i in range(0, num_sink_nodes):
            container_name = f"sink_node_{i}"
            sink_ni = ZdgNode(container_name, "zdg", node_demo_command)
            sink_ni.volumes.append(f"{os.path.split(__file__)[0]}:{working_dir}")
            sink_node_list.append(sink_ni)
            print(f"Creating sink node {container_name}")

        # Connecting nodes
        zmq_port = 5550
        port_data = {}
        for source_ni in source_node_list:
            assert isinstance(source_ni, ZdgNode)
            container_name = source_ni.container_name

            inbound_list_h = []
            inbound_list_p = []

            outbound_list_h = [n.container_name for n in middle_node_list]
            outbound_list_p = []
            for h in outbound_list_h:
                zmq_port += 1
                outbound_list_p.append(zmq_port)
                port_data[f"{container_name}_to_{h}"] = zmq_port

            source_ni.update_inbound_list(inbound_list_h, inbound_list_p)
            source_ni.update_outbound_list(outbound_list_h, outbound_list_p)
            compose_data["services"][container_name] = source_ni.update_yml()[container_name]
            compose_data["services"][container_name]["working_dir"] = working_dir

            print(f"Connecting source node {container_name} to speak to:")
            for h, p in zip(outbound_list_h, outbound_list_p):
                print(f"  middle node {h} in port {p}")

        for middle_ni in middle_node_list:
            assert isinstance(middle_ni, ZdgNode)
            container_name = middle_ni.container_name

            inbound_list_h = [n.container_name for n in source_node_list]
            inbound_list_p = []
            for k, v in port_data.items():
                _ = v
                if f"_to_{container_name}" in k:
                    inbound_list_p.append(port_data[k])

            outbound_list_h = [n.container_name for n in sink_node_list]
            outbound_list_p = []
            for h in outbound_list_h:
                zmq_port += 1
                outbound_list_p.append(zmq_port)
                port_data[f"{container_name}_to_{h}"] = zmq_port

            middle_ni.update_inbound_list(inbound_list_h, inbound_list_p)
            middle_ni.update_outbound_list(outbound_list_h, outbound_list_p)
            compose_data["services"][container_name] = middle_ni.update_yml()[container_name]
            compose_data["services"][container_name]["working_dir"] = working_dir

            print(f"Connecting middle node {container_name} to speak to:")
            for h, p in zip(outbound_list_h, outbound_list_p):
                print(f"  middle node {h} in port {p}")

        for sink_ni in sink_node_list:
            assert isinstance(sink_ni, ZdgNode)
            container_name = sink_ni.container_name

            inbound_list_h = [n.container_name for n in middle_node_list]
            inbound_list_p = []
            for k, v in port_data.items():
                _ = v
                if f"_to_{container_name}" in k:
                    inbound_list_p.append(port_data[k])

            outbound_list_h = []
            outbound_list_p = []

            sink_ni.update_inbound_list(inbound_list_h, inbound_list_p)
            sink_ni.update_outbound_list(outbound_list_h, outbound_list_p)
            compose_data["services"][container_name] = sink_ni.update_yml()[container_name]
            compose_data["services"][container_name]["working_dir"] = working_dir

            print(f"Connecting sink node {container_name} to speak to:")
            for h, p in zip(outbound_list_h, outbound_list_p):
                print(f"  sink node {h} in port {p}")

        print("port_data")
        for k, v in port_data.items():
            print(f"Comunication from {k} uses port {v}")

        u_yaml_path = str(current_dir / "compose_dgraph.yml")
        ZdgNode.write_yml(compose_data, u_yaml_path)
        u_lines = ["\n"]
        for l_i in ZdgNode.get_comments():
            u_lines.append(l_i)
            u_lines.append("\n")
        with open(u_yaml_path, "a", encoding="utf-8") as f_d:
            f_d.writelines(u_lines)


if __name__ == "__main__":
    ZdgNode.demo_compose()
